@@ -17,6 +17,7 @@
 #include "QXmppMessageReaction.h"
 #include "QXmppMixInvitation.h"
 
+#include "Algorithms.h"
 #include "StringLiterals.h"
 #ifdef BUILD_OMEMO
 #include "QXmppOmemoElement_p.h"
@@ -27,12 +28,16 @@
 #include "QXmppUtils.h"
 #include "QXmppUtils_p.h"
 
+#include <ranges>
+
 #include <QDateTime>
 #include <QDomElement>
 #include <QTextStream>
 #include <QXmlStreamWriter>
 
+using namespace QXmpp;
 using namespace QXmpp::Private;
+namespace views = std::views;
 
 constexpr auto CHAT_STATES = to_array<QStringView>({
     {},
@@ -178,6 +183,9 @@ public:
     QVector<QXmppFileShare> sharedFiles;
     QVector<QXmppFileSourcesAttachment> fileSourcesAttachments;
 
+    // XEP-0461: Message Replies
+    std::optional<Reply> reply;
+
     // XEP-0482: Call Invites
     std::optional<QXmppCallInviteElement> callInviteElement;
 };
@@ -219,16 +227,12 @@ bool QXmppMessage::isXmppStanza() const
 }
 
 /// Returns the message's body.
-
 QString QXmppMessage::body() const
 {
     return d->body;
 }
 
 /// Sets the message's body.
-///
-/// \param body
-
 void QXmppMessage::setBody(const QString &body)
 {
     d->body = body;
@@ -1304,6 +1308,105 @@ void QXmppMessage::setFallbackMarkers(const QVector<QXmppFallback> &fallbackMark
     d->fallbackMarkers = fallbackMarkers;
 }
 
+///
+/// Returns the body or subject text of the message without parts that have a fallback marker
+/// with a supported namespace.
+///
+/// \param element whether to use the body or the subject text
+/// \param supportedNamespaces
+///
+/// \since QXmpp 1.9
+///
+QString QXmppMessage::readFallbackRemovedText(QXmppFallback::Element element, const QVector<QString> &supportedNamespaces) const
+{
+    // filter out all QXmppFallback::Reference s
+    auto markers = d->fallbackMarkers |
+        views::filter([&](const auto &marker) { return contains(supportedNamespaces, marker.forNamespace()); }) |
+        views::transform([](auto &&marker) { return marker.references(); });
+
+    // collect references with correct namespace and element type (body or subject)
+    QVector<QXmppFallback::Range> references;
+    for (const auto &markerReferences : markers) {
+        for (const auto &ref : markerReferences) {
+            if (ref.element == element) {
+                // early exit: element without range means whole text is fallback
+                if (!ref.range) {
+                    return {};
+                }
+                references.push_back(ref.range.value());
+            }
+        }
+    }
+
+    // sort by begin of fallback
+    std::ranges::sort(references, {}, &QXmppFallback::Range::start);
+
+    const auto &fullText = element == QXmppFallback::Subject ? d->subject : d->body;
+    QString output;
+    qsizetype index = 0;
+    for (const auto &range : std::as_const(references)) {
+        if (!(range.start < fullText.size()) || !(0 < range.end <= fullText.size())) {
+            // skip markers with invalid start/begin
+            continue;
+        }
+
+        // fallback marker marks [start, end)
+        // we want to copy the section before the marker: [index, start)
+
+        if (index < range.start) {
+            output.append(QStringView(fullText.data() + index, fullText.data() + range.start));
+        }
+        index = range.end;
+    }
+    // append rest of the string (after the last fallback marker)
+    if (index < fullText.size()) {
+        output.append(QStringView(fullText.data() + index, fullText.data() + fullText.size()));
+    }
+
+    return output;
+}
+
+///
+/// Returns the parts of the body or subject that are marked as fallback for this namespace.
+///
+/// \param element whether to use the body or the subject text
+/// \param forNamespace
+///
+/// \since QXmpp 1.9
+///
+QString QXmppMessage::readFallbackText(QXmppFallback::Element element, QStringView forNamespace) const
+{
+    const auto &fullText = element == QXmppFallback::Subject ? d->subject : d->body;
+
+    // filter out all QXmppFallback::Reference s
+    auto markers = d->fallbackMarkers |
+        views::filter([&](const auto &marker) { return marker.forNamespace() == forNamespace; }) |
+        views::transform([](auto &&marker) { return marker.references(); });
+
+    // collect references with correct namespace and element type (body or subject)
+    QVector<QXmppFallback::Range> references;
+    for (const auto &markerReferences : markers) {
+        for (const auto &ref : markerReferences) {
+            if (ref.element == element) {
+                // early exit: element without range means whole text is fallback
+                if (!ref.range) {
+                    return fullText;
+                }
+                references.push_back(ref.range.value());
+            }
+        }
+    }
+
+    // sort by begin of fallback
+    std::ranges::sort(references, {}, &QXmppFallback::Range::start);
+
+    QString output;
+    for (const auto &range : references) {
+        output.append(QStringView(fullText.data() + range.start, fullText.data() + range.end));
+    }
+    return output;
+}
+
 ///
 /// Returns an included trust message element as defined by
 /// \xep{0434, Trust Messages (TM)}.
@@ -1393,6 +1496,47 @@ void QXmppMessage::setFileSourcesAttachments(const QVector<QXmppFileSourcesAttac
     d->fileSourcesAttachments = fileSourcesAttachments;
 }
 
+///
+/// Returns the message reply extension as defined in \xep{0461, Message Replies}.
+///
+/// \since QXmpp 1.9
+///
+std::optional<QXmpp::Reply> QXmppMessage::reply() const
+{
+    return d->reply;
+}
+
+///
+/// Sets the message reply extension as defined in \xep{0461, Message Replies}.
+///
+/// \since QXmpp 1.9
+///
+void QXmppMessage::setReply(const std::optional<QXmpp::Reply> &reply)
+{
+    d->reply = reply;
+}
+
+///
+/// Returns the part of the body that is marked as fallback for \xep{0461, Message Replies},
+/// without any quotation marks ('> ').
+///
+/// \since QXmpp 1.9
+///
+QString QXmppMessage::readReplyQuoteFromBody() const
+{
+    auto replyFallbackBody = readFallbackText(QXmppFallback::Body, ns_reply.toString());
+    auto lines = replyFallbackBody.split(u'\n');
+    // remove '> ' quotation
+    for (auto &line : lines) {
+        if (line == u'>') {
+            line = QString();
+        } else if (line.startsWith(u"> ")) {
+            line = line.mid(QStringView(u"> ").size());
+        }
+    }
+    return lines.join(u'\n');
+}
+
 ///
 /// Returns a Call Invite element as defined in \xep{0482, Call Invites}.
 ///
@@ -1547,13 +1691,6 @@ bool QXmppMessage::parseExtension(const QDomElement &element, QXmpp::SceMode sce
             return true;
         }
 #endif
-        // XEP-0428: Fallback Indication
-        if (checkElement(element, u"fallback", ns_fallback_indication)) {
-            if (auto fallback = QXmppFallback::fromDom(element)) {
-                d->fallbackMarkers.push_back(std::move(*fallback));
-            }
-            return true;
-        }
         // XEP-0482: Call Invites
         if (QXmppCallInviteElement::isCallInviteElement(element)) {
             QXmppCallInviteElement callInviteElement;
@@ -1717,6 +1854,14 @@ bool QXmppMessage::parseExtension(const QDomElement &element, QXmpp::SceMode sce
             }
             return true;
         }
+        // XEP-0461: Message Replies
+        if (checkElement(element, u"reply", ns_reply)) {
+            d->reply = Reply {
+                element.attribute(u"to"_s),
+                element.attribute(u"id"_s),
+            };
+            return true;
+        }
         if (checkElement(element, u"sources", ns_sfs)) {
             if (auto fileSources = QXmppFileSourcesAttachment::fromDom(element)) {
                 d->fileSourcesAttachments.push_back(std::move(*fileSources));
@@ -1724,6 +1869,16 @@ bool QXmppMessage::parseExtension(const QDomElement &element, QXmpp::SceMode sce
             return true;
         }
     }
+
+    // read from both public and private extensions:
+
+    // XEP-0428: Fallback Indication
+    if (checkElement(element, u"fallback", ns_fallback_indication)) {
+        if (auto fallback = QXmppFallback::fromDom(element)) {
+            d->fallbackMarkers.push_back(std::move(*fallback));
+        }
+        return true;
+    }
     return false;
 }
 
@@ -1801,11 +1956,6 @@ void QXmppMessage::serializeExtensions(QXmlStreamWriter *writer, QXmpp::SceMode
             d->omemoElement->toXml(writer);
         }
 #endif
-
-        // XEP-0428: Fallback Indication
-        for (const auto &fallback : d->fallbackMarkers) {
-            fallback.toXml(writer);
-        }
     }
 
     if (sceMode & QXmpp::SceSensitive) {
@@ -1985,11 +2135,29 @@ void QXmppMessage::serializeExtensions(QXmlStreamWriter *writer, QXmpp::SceMode
             fileSources.toXml(writer);
         }
 
+        // XEP-0461: Message Replies
+        if (d->reply) {
+            writer->writeStartElement(QSL65("reply"));
+            writer->writeDefaultNamespace(toString65(ns_reply));
+            writeOptionalXmlAttribute(writer, u"to", d->reply->to);
+            writer->writeAttribute(QSL65("id"), d->reply->id);
+            writer->writeEndElement();
+        }
+
         // XEP-0482: Call Invites
         if (d->callInviteElement) {
             d->callInviteElement->toXml(writer);
         }
     }
+
+    // serialize in private and in public part
+
+    // XEP-0428: Fallback Indication
+    // fallback markers may be used in the private part (e.g. message replies) but also in the
+    // public part (e.g. the fallback body for e2ee messages)
+    for (const auto &fallback : d->fallbackMarkers) {
+        fallback.toXml(writer);
+    }
 }
 
 struct QXmppFallbackPrivate : QSharedData {
